/*
 * This file is part of the Colobot: Gold Edition source code
 * Copyright (C) 2001-2023, Daniel Roux, EPSITEC SA & TerranovaTeam
 * http://epsitec.ch; http://colobot.info; http://github.com/colobot
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://gnu.org/licenses
 */

#pragma once


#include "object/motion/motion.h"


class CMotionVehicle : public CMotion
{
public:
    CMotionVehicle(COldObject* object);
    ~CMotionVehicle();

    void        DeleteObject(bool bAll=false) override;
    void        Create(Math::Vector pos, float angle, ObjectType type, float power, Gfx::COldModelManager* modelManager) override;
    bool        EventProcess(const Event &event) override;

protected:
    void        CreatePhysics(ObjectType type);
    bool        EventFrame(const Event &event);
    bool        EventFrameFly(const Event &event);
    bool        EventFrameInsect(const Event &event);
    bool        EventFrameCanoni(const Event &event);
    void        UpdateTrackMapping(float left, float right, ObjectType type);

protected:
    float       m_wheelTurn[4];
    float       m_flyPaw[3];
    float       m_posTrackLeft;
    float       m_posTrackRight;
    int         m_partiReactor;
    float       m_armTimeAbs;
    float       m_armMember;
    float       m_canonTime;
    Math::Vector    m_wheelLastPos;
    Math::Vector    m_wheelLastAngle;
    Math::Vector    m_posKey;
    bool        m_bFlyFix;
};
